/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)
    
    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 
*/

#include "config.h"
#include "virt.h"

#undef PREF

#ifdef HAVE___OPEN
#define PREF(x) __ ## x
#else 
#define PREF(x) _ ## x
#endif

#undef VIRT
#define VIRT(name) __av_virt ## name

#undef PATH_T
#define PATH_T const char *


#define ALIAS1(name, alias, rettyp, a1t) \
extern rettyp alias (a1t);               \
rettyp alias (a1t a1)                    \
{ return VIRT(name) (a1); }

#define ALIAS2(name, alias, rettyp, a1t, a2t) \
extern rettyp alias (a1t, a2t);               \
rettyp alias (a1t a1, a2t a2)                 \
{ return VIRT(name) (a1, a2); }

#define ALIAS3(name, alias, rettyp, a1t, a2t, a3t) \
extern rettyp alias (a1t, a2t, a3t);               \
rettyp alias (a1t a1, a2t a2, a3t a3)              \
{ return VIRT(name) (a1, a2, a3); }

#define ALIAS4(name, alias, rettyp, a1t, a2t, a3t, a4t) \
extern rettyp alias (a1t, a2t, a3t, a4t);               \
rettyp alias (a1t a1, a2t a2, a3t a3, a4t a4)           \
{ return VIRT(name) (a1, a2, a3, a4); }

#define ALIAS6(name, alias, rettyp, a1t, a2t, a3t, a4t, a5t, a6t) \
extern rettyp alias (a1t, a2t, a3t, a4t, a5t, a6t);               \
rettyp alias (a1t a1, a2t a2, a3t a3, a4t a4, a5t a5, a6t a6)     \
{ return VIRT(name) (a1, a2, a3, a4, a5, a6); }


ALIAS3(open, open, int, PATH_T, int, int)
ALIAS3(open, PREF(open), int, PATH_T, int, int)

ALIAS2(creat, creat, int, PATH_T, mode_t)
ALIAS2(creat, PREF(creat), int, PATH_T, mode_t)

ALIAS1(close, close, int, int)
ALIAS1(close, PREF(close), int, int)

ALIAS3(read, read, ssize_t, int, void *, size_t)
ALIAS3(read, PREF(read), ssize_t, int, void *, size_t)

ALIAS3(write, write, ssize_t, int, const void *, size_t)
ALIAS3(write, PREF(write), ssize_t, int, const void *, size_t)

ALIAS3(lseek, lseek, off_t, int, off_t, int)
ALIAS3(lseek, PREF(lseek), off_t, int, off_t, int)


#ifdef HAVE__XSTAT

#include <errno.h>

#ifdef LINUX_GLIBC
#define STAT_VERSION 3
#else
#define STAT_VERSION 1
#endif

#define XSTATALIAS(name, alias, a2t)                     \
extern int alias (int, a2t, void *);                     \
int alias(int ver, a2t a2, void *buf)                    \
{ if(ver != STAT_VERSION) { errno = EINVAL; return -1; } \
  return VIRT(name) (a2, buf); }

XSTATALIAS(fstat, _fxstat, int)
XSTATALIAS(fstat, __fxstat, int)

XSTATALIAS(stat, _xstat, PATH_T)
XSTATALIAS(stat, __xstat, PATH_T)

XSTATALIAS(lstat, _lxstat, PATH_T)
XSTATALIAS(lstat, __lxstat, PATH_T)

#define MKNOD_VERSION 1

extern int _xmknod(int, const char *, mode_t, dev_t *);
int _xmknod(int ver, const char *path, mode_t mode, dev_t *devp)
{ if(ver != MKNOD_VERSION) {errno = EINVAL; return -1; }
  return VIRT(mknod) (path, mode, *devp); }

extern int __xmknod(int, const char *, mode_t, dev_t *);
int __xmknod(int ver, const char *path, mode_t mode, dev_t *devp)
{ if(ver != MKNOD_VERSION) {errno = EINVAL; return -1; }
  return VIRT(mknod) (path, mode, *devp); }

#else /* HAVE__XSTAT */

ALIAS2(fstat, fstat, int, int, void *)
ALIAS2(fstat, PREF(fstat), int, int, void *)

ALIAS2(stat, stat, int, PATH_T, void *)
ALIAS2(stat, PREF(stat), int, PATH_T, void *)

ALIAS2(lstat, lstat, int, PATH_T, void *)
ALIAS2(lstat, PREF(lstat), int, PATH_T, void *)

ALIAS3(mknod, mknod, int, PATH_T, mode_t, dev_t)
ALIAS3(mknod, PREF(mknod), int, PATH_T, mode_t, dev_t)

#endif /* HAVE__XSTAT */


ALIAS3(readlink, readlink, int, PATH_T, char *, size_t)
ALIAS3(readlink, PREF(readlink), int, PATH_T, char *, size_t)

ALIAS2(access, access, int, PATH_T, int)
ALIAS2(access, PREF(access), int, PATH_T, int)

ALIAS1(chdir, chdir, int, PATH_T)
ALIAS1(chdir, PREF(chdir), int, PATH_T)

ALIAS1(fchdir, fchdir, int, int)
ALIAS1(fchdir, PREF(fchdir), int, int)

ALIAS3(getdents, getdents, int, int, void *, size_t)
ALIAS3(getdents, PREF(getdents), int, int, void *, size_t)

ALIAS3(execve, execve, int, PATH_T, char *const *, char *const *)
ALIAS3(execve, PREF(execve), int, PATH_T, char *const *, char *const *)

ALIAS2(link, link, int, PATH_T, PATH_T)
ALIAS2(link, PREF(link), int, PATH_T, PATH_T)

ALIAS1(unlink, unlink, int, PATH_T)
ALIAS1(unlink, PREF(unlink), int, PATH_T)

ALIAS1(rmdir, rmdir, int, PATH_T)
ALIAS1(rmdir, PREF(rmdir), int, PATH_T)

ALIAS2(mkdir, mkdir, int, PATH_T, mode_t)
ALIAS2(mkdir, PREF(mkdir), int, PATH_T, mode_t)

ALIAS2(symlink, symlink, int, PATH_T, PATH_T)
ALIAS2(symlink, PREF(symlink), int, PATH_T, PATH_T)

ALIAS2(rename, rename, int, PATH_T, PATH_T)
ALIAS2(rename, PREF(rename), int, PATH_T, PATH_T)

ALIAS2(chmod, chmod, int, PATH_T, mode_t)
ALIAS2(chmod, PREF(chmod), int, PATH_T, mode_t)

ALIAS2(fchmod, fchmod, int, int, mode_t)
ALIAS2(fchmod, PREF(fchmod), int, int, mode_t)

ALIAS3(chown, chown, int, PATH_T, uid_t, gid_t)
ALIAS3(chown, PREF(chown), int, PATH_T, uid_t, gid_t)

ALIAS3(fchown, fchown, int, int, uid_t, gid_t)
ALIAS3(fchown, PREF(fchown), int, int, uid_t, gid_t)

ALIAS2(utime, utime, int, PATH_T, void *)
ALIAS2(utime, PREF(utime), int, PATH_T, void *)

#ifdef EMULATE_UTIMES
ALIAS2(utimes, utimes, int, PATH_T, void *)
ALIAS2(utimes, PREF(utimes), int, PATH_T, void *)
#endif

#ifdef HAVE_LCHOWN
ALIAS3(lchown, lchown, int, PATH_T, uid_t, gid_t)
ALIAS3(lchown, PREF(lchown), int, PATH_T, uid_t, gid_t)
#endif


#ifdef HAVE_MMAP

ALIAS6(mmap, mmap, void *, void *, size_t, int, int, int, off_t)
ALIAS6(mmap, PREF(mmap), void *, void *, size_t, int, int, int, off_t)

#endif /* HAVE_MMAP */


#ifdef HAVE_ACL

ALIAS4(acl, acl, int, PATH_T, int, int, void *)

#endif /* HAVE_ACL */

#ifdef HAVE_OPEN64

ALIAS3(open64, open64, int, PATH_T, int, int)
ALIAS3(open64, PREF(open64), int, PATH_T, int, int)

ALIAS2(creat64, creat64, int, PATH_T, mode_t)
ALIAS2(creat64, PREF(creat64), int, PATH_T, mode_t)

ALIAS3(lseek64, llseek, offset_t, int, offset_t, int)
#if 0
ALIAS3(lseek64, PREF(llseek), offset_t, int, offset_t, int)
#endif

ALIAS3(lseek64, lseek64, offset_t, int, offset_t, int)
ALIAS3(lseek64, PREF(lseek64), offset_t, int, offset_t, int)

ALIAS2(fstat64, fstat64, int, int, void *)
ALIAS2(fstat64, PREF(fstat64), int, int, void *)

ALIAS2(stat64, stat64, int, PATH_T, void *)
ALIAS2(stat64, PREF(stat64), int, PATH_T, void *)

ALIAS2(lstat64, lstat64, int, PATH_T, void *)
ALIAS2(lstat64, PREF(lstat64), int, PATH_T, void *)

ALIAS3(getdents64, getdents64, int, int, void *, size_t)
ALIAS3(getdents64, PREF(getdents64), int, int, void *, size_t)

ALIAS6(mmap64, mmap64, void *, void *, size_t, int, int, int, offset_t)
#if 0
ALIAS6(mmap64, PREF(mmap64), void *, void *, size_t, int, int, int, offset_t)
#endif

#endif /* HAVE_OPEN64 */

#ifdef EMULATE_DIROPS

ALIAS1(opendir, opendir, void *, PATH_T)
ALIAS1(closedir, closedir, int, void *)
ALIAS1(readdir, readdir, void *, void *)
ALIAS1(telldir, telldir, off_t, void *)
ALIAS2(seekdir, seekdir, void, void *, off_t)
ALIAS1(rewinddir, rewinddir, void, void *)

#endif /* EMULATE_DIROPS */
