#include <stdio.h>
#include <stdlib.h>

/* #define HAVE_LIBC_MALLOC */

#define SEGV() (*(char *)0 = 1)
/* #define SEGV() (_exit(1)) */


#ifndef HAVE_LIBC_MALLOC

#ifdef __BORLANDC__
#define MEMSIZE 32768
#else
#define MEMSIZE 16777216
#endif

typedef unsigned char byte;
static byte xmem[MEMSIZE];

static int xmeminit = 0;

static int xat = 0;
static int xmused = 0;

#define HOV 8

#define NEXT(ix) (*((int *)(xmem+(ix)+4)))
#define PREV(ix) (*((int *)(xmem+(ix))))
#define MEMP(ix) (xmem+(ix)+HOV)
#define PTIX(ptr) ((int) (ptr) - (int) (xmem) - HOV)



#define USEDBIT 0x01

#define ROUND 0x07

#define SIZE(bd) ((bd) & ~ROUND)
#define USED(bd) ((bd) & USEDBIT)

#define NEXTIX(ix, size) (((ix)+(size)+HOV) & (MEMSIZE-1))
#define PREVIX(ix, size) (((ix)-(size)-HOV) & (MEMSIZE-1))

#define MIX(ptr) ((int) (ptr) - (int) xmem)

#define ILLIX(ix) ((ix) < 0 || (ix) >= MEMSIZE || ((ix) & ROUND) != 0)

#define checkinit() if(!xmeminit) init_xmem()

static void init_xmem(void)
{
  xmeminit = 1;
  NEXT(0) = MEMSIZE - HOV;
  PREV(0) = MEMSIZE - HOV;
}

static int checknext(int ix, int *nsizep)
{
  register int nsize, nix, nbd, pbd;

  nbd = NEXT(ix);
  nsize = SIZE(nbd);
  nix = NEXTIX(ix, nsize);
  pbd = PREV(nix);

  if(nbd != pbd) {
    fprintf(stderr,
            "XMEM: Memory error at %i: nextdesc (%i) != prevdesc (%i)\n",
            ix, nbd, pbd);
    SEGV();
    return -1;
  }

  *nsizep = nsize;

  return nix;
}

static int findmem(size_t size)
{
  int nat;

  nat = xat;
  do {
    int nsize, nix;

    nix = checknext(nat, &nsize);
    if(!USED(NEXT(nat)) && nsize >= size) return nat;
    nat = nix;

  } while(nat != xat);

  fprintf(stderr, "XMEM: Out of memory (total: %i, used: %i)\n",
          MEMSIZE, xmused);

  return -1;
}

static int splitmem(int ix, size_t size)
{
  int osize, size1, size2;
  int nix, eix;

  eix = checknext(ix, &osize);
  size1 = (size + ROUND) & ~ROUND;
  if(osize - size1 < 24) return osize;

  size2 = osize - size1 - HOV;

  nix = NEXTIX(ix, size1);

  NEXT(ix) = size1;
  PREV(nix) = size1;
  NEXT(nix) = size2;
  PREV(eix) = size2;

  return size1;
}


static int unsplit(int ix)
{
  int size1, size2, nsize;
  int nix, eix;

  nix = checknext(ix, &size1);
  if(xat == nix || nix == 0) return size1;

  eix = checknext(nix, &size2);

  nsize = size1 + size2 + HOV;

  NEXT(ix) = nsize;
  PREV(eix) = nsize;

  return nsize;
}


static int allocmem(size_t size)
{
  int ix;
  int nsize;

  ix = findmem(size);
  if(ix < 0) return -1;

  nsize = splitmem(ix, size);

  NEXT(ix) |= USEDBIT;
  xat = NEXTIX(ix, nsize);
  PREV(xat) |= USEDBIT;

  xmused += (nsize + HOV);

  return ix;
}

static int freemem(int ix)
{
  int nsize;
  int nix, pix;

  nix = checknext(ix, &nsize);

  NEXT(ix) &= ~USEDBIT;
  PREV(nix) &= ~USEDBIT;

  pix = PREVIX(ix, SIZE(PREV(ix)));

  if(!USED(NEXT(nix))) unsplit(ix);
  if(!USED(PREV(ix))) unsplit(pix);

  xmused -= (nsize + HOV);

  return 1;
}

/* --------------------------------------- *
 *     xmalloc                             *
 *     xfree                               *
 *     xrealloc                            *
 * --------------------------------------- */

static void *xmalloc(size_t size)
{
  int mi;

  checkinit();

  if(size == 0) size = 1;

  mi = allocmem(size);
  if(mi < 0) return NULL;

  return (void *) MEMP(mi);
}

static void xfree(void *ptr)
{
  int mi;

  checkinit();

  if(ptr == NULL) return;
  mi = PTIX(ptr);

  if(ILLIX(mi)) {
    fprintf(stderr, "XMEM: free - illegal pointer\n");
    SEGV();
    return;
  }

  freemem(mi);
}

static void *xrealloc(void *ptr, size_t size)
{
  byte *np, *op;
  int nsize, i;
  int mi, ix;

  checkinit();

  if(size == 0) {
    xfree(ptr);
    return NULL;
  }

  if(ptr == NULL) return xmalloc(size);

  mi = PTIX(ptr);
  if(ILLIX(mi)) {
    fprintf(stderr, "XMEM: free - illegal pointer\n");
    SEGV();
    return NULL;
  }

  checknext(mi, &nsize);

  if(size <= nsize) return ptr;

  ix = allocmem(size);
  if(ix < 0) return NULL;

  np = MEMP(ix);
  op = (byte *) ptr;

  for(i = nsize; i; np++, op++, i--) *np = *op;

  freemem(mi);

  return MEMP(ix);
}


#else /* HAVE_LIBC_MALLOC */

#define xmalloc  __libc_malloc
#define xrealloc __libc_realloc
#define xfree    __libc_free

extern void *__libc_malloc(size_t size);
extern void *__libc_realloc(void *ptr, size_t size);
extern void  __libc_free(void *ptr);

#endif /* HAVE_LIBC_MALLOC */


/* ---------------------------------------------------------------------- *
 *                                                                        *
 *                           End of XMEM                                  *
 *                                                                        *
 * ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- *
 *                                                                        *
 *                          Amalloc follows                               *
 *                                                                        *
 * ---------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#ifdef DEFAULT_SHOW_UNFREED
#define SHOW_UNFREED 1
#else
#define SHOW_UNFREED 0
#endif

typedef struct {
  char *ptr;
  size_t len;
  int seed;
  int serial;
  short id;
  short freed;
} Minfo;

static int protb = 16, prota = 16;
static int inited = 0;
static Minfo *mis;
static int mnum;
static int mrnum;
static int missiz;
static int serial;
static int show_unfreed = SHOW_UNFREED;

static int iastopat = -1;

#define MISINC 16

#define SSIZ sizeof(size_t)


#define notinited() if(!inited) init_amalloc()

extern void *_iamalloc(size_t size, int id);
extern void *_iarealloc(void *ptr, size_t size, int id);
extern void  _iafree(void *ptr, int id);
extern void amalloc_show_unfreed(int id);

static int insert_entry(char *ptr, size_t len, int seed, int id)
{
  int i;

  if(mnum == missiz) {
    missiz += MISINC;
    mis = (Minfo *) xrealloc(mis, sizeof(Minfo) * missiz);
    if(mis == NULL) return -1;
  }

  i = mnum;

  mis[i].ptr = ptr;
  mis[i].len = len;
  mis[i].seed = seed;
  mis[i].id = id;
  mis[i].serial = serial++;
  mis[i].freed = 0;

  mnum++;
  mrnum++;

  return i;
}

static void delete_entry(int i)
{

  memmove(mis+i, mis+i+1, sizeof(Minfo)*(mnum-(i+1)));
  mnum--;

  mrnum--;

  /*
  mis[i].freed = 1;
  */

}


static int find_entry(char *ptr)
{
  int i;

  for(i = 0; i < mnum; i++) if(!mis[i].freed && mis[i].ptr == ptr) return i;
  return -1;
}

static int check_mem(int i, int tofree)
{
  int j;
  int err;
  char *p;

  srand(mis[i].seed);
  p = mis[i].ptr-protb;
  err = 0;

  for(j = protb; j; p++, j--)
    if(*p != (char) rand()) {
      err = 1;
      fprintf(stderr,
              "amalloc: protected byte changed (%s) %p/%i#%04i-%i(%c/%i)\n",
              mis[i].freed ? "close" : (tofree ? "free" : "realloc"),
              mis[i].ptr, mis[i].len, mis[i].serial, j, *p, *p);

    }

  if(mis[i].freed) {
    for(j = mis[i].len; j; p++, j--)
      if(*p != (char) rand()) {
        err = 1;
        fprintf(stderr,
                "amalloc: protected byte changed (%s) %p/%i#%04i@%i(%c/%i)\n",
                mis[i].freed ? "close" : (tofree ? "free" : "realloc"),
                mis[i].ptr, mis[i].len, mis[i].serial, mis[i].len - j, *p, *p);
      }
  }
  else {
    if(!tofree) for(j = mis[i].len; j; p++, j--) rand();
    else for(j = mis[i].len; j; p++, j--) *p = rand();
  }

  for(j = prota; j; p++, j--)
    if(*p != (char) rand()) {
      err = 1;
      fprintf(stderr,
              "amalloc: protected byte changed (%s) %p/%i#%04i+%i(%c/%i)\n",
              mis[i].freed ? "close" : (tofree ? "free" : "realloc"),
              mis[i].ptr, mis[i].len, mis[i].serial, prota - j, *p, *p);

    }
  if(err) {
    SEGV();
  }

  return 0;
}

static int check_entry(char *ptr, int tofree, int id)
{
  int i;

  i = find_entry(ptr);

  if(i < 0) {
    fprintf(stderr,
            "amalloc: freeing or reallocing illegal pointer %p\n", ptr);
    SEGV();
    return -1;
  }

  if(id != mis[i].id) {
    fprintf(stderr,
            "amalloc: freeing and allocation differ %p/%i#%04i (id: %i-%i)\n",
            mis[i].ptr, mis[i].len, mis[i].serial, mis[i].id, id);
    SEGV();
    return -1;
  }

  check_mem(i, tofree);

  return i;
}



static void close_amalloc(void)
{
  int i;
  int shunf;

  /*
  for(i = 0; i < mnum; i++) {
    check_mem(i, 0);
    if(mis[i].freed) {
      fprintf(stderr, "freeing %i\n", i);
      xfree(mis[i].ptr - protb);
    }
  }
  */

  if(mrnum > 0) {
    shunf = 0;
    if(show_unfreed) 
      for(i = 0; i < mnum; i++) 
	if(!mis[i].freed && mis[i].id >= show_unfreed) {
	  shunf = 1;
	}
    
    if(shunf) fprintf(stderr, "amalloc: Unfreed pointers remaining:\n");
    for(i = 0; i < mnum; i++)
      if(!mis[i].freed) {
	if(show_unfreed && mis[i].id >= show_unfreed) 
	  fprintf(stderr, "#%04i %p/%i\n",
		  mis[i].serial,  mis[i].ptr, mis[i].len);
        check_mem(i, 0);
      }
    if(shunf) fprintf(stderr, "\n");
  }
  xfree(mis);
  inited = -1;
}

void amalloc_show_unfreed(int id)
{
  show_unfreed = id;
}

static void init_amalloc(void)
{
  inited = 1;

  mis = NULL;
  mnum = 0;
  missiz = 0;
  serial = 0;

  atexit(close_amalloc);
}

static void mat(void)
{
  fprintf(stderr, "stopped at %i\n", serial);
}

void *_iamalloc(size_t size, int id)
{
  char *me, *p, *ptr;
  int i, j;


  if(inited < 0) {
    /*    fprintf(stderr, "amalloc: called after close\n"); */
    return NULL;
  }

  notinited();

  if(iastopat == serial) mat();

  /*
  if(size < 0) {
    fprintf(stderr, "amalloc: illegal size: %i\n", size);
    SEGV();
  }
  */
  if(size == 0) fprintf(stderr, "amalloc: Warning: size == 0\n");

  me = (char *) xmalloc(size + protb + prota);
  if(me == NULL) return NULL;

  ptr = me+protb;

  i = insert_entry(ptr, size, time(NULL), id);
  if(i < 0) {
    xfree(me);
    fprintf(stderr, "amalloc: Unable to allocate space\n");
    return NULL;
  }

  srand(mis[i].seed);
  p = me;
  for(j = protb+size+prota; j; p++, j--) *p = (char) rand();

  return (void *) ptr;
}

void _iafree(void *ip, int id)
{
  char *ptr, *me;
  int i;

  if(inited < 0) {
    /*     fprintf(stderr, "afree: called after close\n"); */
    return;
  }

  notinited();


  ptr = (char *) ip;

  if(ptr == NULL) return;

  i = check_entry(ptr, 1, id);

  delete_entry(i);

  me = ptr - protb;
  xfree(me);
}

void *_iarealloc(void *ip, size_t size, int id)
{
  char *me, *p, *ptr;
  int i,j;

  if(inited < 0) {
    /*    fprintf(stderr, "arealloc: called after close\n"); */
    return NULL;
  }

  notinited();

  if(size == 0) {
    _iafree(ip, id);
    return NULL;
  }

  ptr = (char *) ip;

  if(ptr == NULL) return _iamalloc(size, id);

  i = check_entry(ptr, 0, id);

  me = (char *) xrealloc(ptr-protb, size + protb + prota);
  if(me == NULL) return NULL;

  ptr = me+protb;

  srand(mis[i].seed);
  p = me;
  for(j = protb; j; p++, j--) rand(); /* no need to rewrite memory */

  if(mis[i].len <= size) {
    for(j = mis[i].len; j; p++, j--) rand();
    for(j = size - mis[i].len; j; p++, j--) *p = (char) rand();
  }
  else for(j = size; j; p++, j--) rand();

  for(j = prota; j; p++, j--) *p = (char) rand();

  mis[i].ptr = ptr;
  mis[i].len = size;

  return (void *) ptr;
}

#if 0

extern int stopat(int i);

int stopat(int i)
{
  iastopat = i;
  if(serial <= i) return 1;
  else return 0;
}


void free(void *ptr)
{
  _iafree(ptr, 4);
}

void *malloc(size_t size)
{
  return _iamalloc(size, 4);
}

void *calloc(size_t nmemb, size_t size)
{
  void *ptr;
  char *p;
  size_t rsize, i;

  rsize = nmemb * size;

  ptr = _iamalloc(rsize, 4);
  if(ptr == NULL) return NULL;

  p = (char *) ptr;
  for(i = rsize; i; p++, i--) *p = '\0';

  return ptr;
}

void *realloc(void *ptr, size_t size)
{
  return _iarealloc(ptr, size, 4);
}
#endif
